from .base import Block, BaseTableBlock
from .text import TextBlock
from .image import ImageBlock
from .spacer import SpacerBlock
from .table import TableBlock
from .container import ContainerBlock
from .row import RowBlock
from .columns import ColumnsBlock
from .grid import GridBlock

from .icon import IconBlock
from .lucide import LucideIconBlock



from .document import Document, DocumentBuilder
from .page import Page

from .styles import ContainerStyle, TextStyle, IconStyle, RenderContext
import os

from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont

def load_fonts():
    font_dir = os.path.join(os.path.dirname(__file__), 'fonts')
    if not os.path.exists(font_dir):
        print(f"Warning: Font directory not found: {font_dir}")
        return

    for f in os.listdir(font_dir):
        if f.lower().endswith('.ttf'):
            font_name = os.path.splitext(f)[0]
            font_path = os.path.join(font_dir, f)
            try:
                pdfmetrics.registerFont(TTFont(font_name, font_path))
            except Exception as e:
                print(f"Failed to register font {font_name}: {e}")

load_fonts()



from dataclasses import dataclass
from typing import Optional
from .base import Style

@dataclass
class ContainerStyle(Style):
    bg_color: Optional[str] = None
    border_color: Optional[str] = None
    border_width: Optional[float] = None
    border_radius: Optional[int] = None
    padding: Optional[int] = None
    margin: Optional[int] = None
    width: Optional[float] = None
    height: Optional[float] = None



from typing import List, Optional
from reportlab.platypus import Table, TableStyle
from .base import Block, BaseTableBlock
from gen_pdf_abc_xyz_will_change_later.styles.context import RenderContext

class ColumnsBlock(BaseTableBlock):
    def __init__(self, children: List[Block], col_widths=None, style=None):
        super().__init__(style=style)
        self.children = children
        self.col_widths = col_widths
        
        self.margin = getattr(style, 'margin', 0) or 0 if style else 0
        self.padding = getattr(style, 'padding', 0) or 0 if style else 0

    def render(self, context: Optional['RenderContext'] = None):
        row_data = []
        for child in self.children:
             r = child.render(context)
             if isinstance(r, list):
                 row_data.append(r)
             else:
                 row_data.append([r])
        
        data = [row_data]
        t = Table(data, colWidths=self.col_widths)
        
        style_cmds = self._get_base_style(context)
        t.setStyle(TableStyle(style_cmds))
        t.spaceBefore = self.margin
        t.spaceAfter = self.margin
        return t



from dataclasses import dataclass
from typing import Optional
from .base import Style

@dataclass
class TextStyle(Style):
    font_name: Optional[str] = None
    font_size: Optional[int] = None
    color: Optional[str] = None
    leading: Optional[int] = None
    alignment: Optional[int] = None
    space_after: Optional[int] = None



from typing import List, Union, Optional
from reportlab.platypus import Table, TableStyle
from reportlab.lib.colors import HexColor
from .base import Block, BaseTableBlock
from gen_pdf_abc_xyz_will_change_later.styles.context import RenderContext

class ContainerBlock(BaseTableBlock):
    def __init__(self, children: Union[Block, List[Block]], style=None):
        super().__init__(style=style)
        self.children = children if isinstance(children, list) else [children]
        
        self.width = getattr(style, 'width', None) if style else None
        self.height = getattr(style, 'height', None) if style else None
        self.margin = getattr(style, 'margin', 0) or 0 if style else 0
        self.padding = getattr(style, 'padding', 0) or 0 if style else 0

    def render(self, context: Optional['RenderContext'] = None):
        if self.style:
            if self.width is None: self.width = getattr(self.style, 'width', None)
            if self.height is None: self.height = getattr(self.style, 'height', None)

        rendered_children = []
        for child in self.children:
            r = child.render(context)
            if isinstance(r, list):
                rendered_children.extend(r)
            else:
                rendered_children.append(r)
        
        data = [[rendered_children]]
        
        col_widths = [self.width] if self.width else None
        row_heights = [self.height] if self.height else None
        
        final_radius = 0
        if self.style:
             final_radius = getattr(self.style, 'border_radius', 0) or 0
             
        if self.height and final_radius > self.height / 2.0:
            final_radius = self.height / 2.0
            
        corner_radii = [final_radius]*4 if final_radius > 0 else None
        
        t = Table(data, colWidths=col_widths, rowHeights=row_heights, cornerRadii=corner_radii)
        
        style_cmds = self._get_base_style(context)
            
        if self.style:
            border_width = getattr(self.style, 'border_width', 0) or 0
            border_color = getattr(self.style, 'border_color', None)
            
            if border_width > 0 and border_color:
                final_border_color = border_color
                c = HexColor(final_border_color) if isinstance(final_border_color, str) else final_border_color
                style_cmds.append(('BOX', (0,0), (-1,-1), border_width, c))

        if final_radius > 0:
             t.setStyle(TableStyle([('ROUNDEDCORNERS', [final_radius]*4)]))
        
        t.setStyle(TableStyle(style_cmds))
        t.spaceBefore = self.margin
        t.spaceAfter = self.margin
        return t



from dataclasses import dataclass
from typing import Dict, Any, Optional

@dataclass
class Style:
    def merge(self, other: 'Style') -> 'Style':
        if not isinstance(other, type(self)):
            return other
        
        new_style = type(self)()
        
        for field in self.__dataclass_fields__:
            v_self = getattr(self, field)
            v_other = getattr(other, field)
            
            if v_other is not None:
                setattr(new_style, field, v_other)
            else:
                setattr(new_style, field, v_self)
                
        return new_style
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        valid_keys = {k for k in cls.__dataclass_fields__}
        filtered = {k: v for k, v in data.items() if k in valid_keys}
        return cls(**filtered)



from typing import List, Any
from reportlab.platypus import Table, TableStyle
from reportlab.lib import colors
from .base import BaseTableBlock

class TableBlock(BaseTableBlock):
    def __init__(self, data: List[List[Any]], col_widths=None, row_heights=None, style=None):
        super().__init__()
        self.data = data
        self.col_widths = col_widths
        self.row_heights = row_heights
        self.style = style or [
            ('GRID', (0, 0), (-1, -1), 1, colors.black),
            ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
            ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
        ]

    def render(self, context=None):
        t = Table(self.data, colWidths=self.col_widths, rowHeights=self.row_heights)
        t.setStyle(TableStyle(self.style))
        return t



from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from gen_pdf_abc_xyz_will_change_later.styles.context import RenderContext
import os
import tempfile
import urllib.request
from reportlab.platypus import Image, Spacer
from reportlab.lib.utils import ImageReader
from .base import Block

class ImageBlock(Block):
    def __init__(self, path: str, style=None):
        self.path = path
        self.style = style
        
        self.width = getattr(style, 'width', None) if style else None
        self.height = getattr(style, 'height', None) if style else None
        self.border_radius = getattr(style, 'border_radius', 0) or 0 if style else 0

    def render(self, context: Optional['RenderContext'] = None):
        if self.style:
             if self.width is None: self.width = getattr(self.style, 'width', None)
             if self.height is None: self.height = getattr(self.style, 'height', None)
             if self.border_radius == 0: 
                 radius = getattr(self.style, 'border_radius', 0)
                 self.border_radius = radius if radius is not None else 0

        image_path = self.path
        
        if self.path.startswith('http://') or self.path.startswith('https://'):
            try:
                import hashlib
                fname = hashlib.md5(self.path.encode()).hexdigest() + ".jpg"
                temp_path = os.path.join(tempfile.gettempdir(), fname)
                
                if not os.path.exists(temp_path):
                    with urllib.request.urlopen(self.path) as response:
                        data = response.read()
                    with open(temp_path, 'wb') as f:
                        f.write(data)
                
                image_path = temp_path
            except Exception as e:
                print(f"Failed to download image: {e}")
                return Spacer(0, 0)

        if self.border_radius > 0:
            try:
                from PIL import Image as PILImage, ImageDraw, ImageOps
                
                with PILImage.open(image_path) as im:
                    im = im.convert("RGBA")
                    
                    mask = PILImage.new('L', im.size, 0)
                    draw = ImageDraw.Draw(mask)
                    
                    target_w_pts = self.width
                    target_h_pts = self.height
                    
                    if target_w_pts and target_h_pts:
                        scale = 3 
                        px_w = int(target_w_pts * scale)
                        px_h = int(target_h_pts * scale)
                        px_radius = int(self.border_radius * scale)
                        
                        im = im.resize((px_w, px_h), PILImage.LANCZOS)
                        
                        mask = PILImage.new('L', (px_w, px_h), 0)
                        draw = ImageDraw.Draw(mask)
                        draw.rounded_rectangle([(0,0), (px_w, px_h)], radius=px_radius, fill=255)
                        
                        im.putalpha(mask)
                        
                        fd, new_path = tempfile.mkstemp(suffix='.png')
                        os.close(fd)
                        im.save(new_path, format="PNG")
                        image_path = new_path
                        
            except ImportError:
                print("Pillow not installed, skipping border radius")
            except Exception as e:
                print(f"Failed to apply border radius: {e}")

        if (self.width is None or self.height is None) and self.width != self.height:
            try:
                utils = ImageReader(image_path)
                iw, ih = utils.getSize()
                aspect = ih / float(iw)
                
                if self.width and self.height is None:
                    self.height = self.width * aspect
                elif self.height and self.width is None:
                    self.width = self.height / aspect
            except Exception:
                pass

        return Image(image_path, width=self.width, height=self.height)



from dataclasses import dataclass, field
from typing import List, Optional, Deque
from collections import deque

from .base import Style
from .box import ContainerStyle
from .text import TextStyle
from .icon import IconStyle

class RenderContext:
    def __init__(self):
        self.style_stack: Deque[Style] = deque()
        self._root_style = ContainerStyle()

    def push_style(self, style: Style):
        self.style_stack.append(style)

    def pop_style(self):
        if self.style_stack:
            self.style_stack.pop()

    def resolve_style(self, base_style: Optional[Style] = None) -> Style:
        final = type(base_style)() if base_style else ContainerStyle()
        
        target_type = type(final)
        
        for s in self.style_stack:
            if isinstance(s, target_type):
                final = final.merge(s)
            
        if base_style:
            final = final.merge(base_style)
            
        return final

    def get_color(self, name_or_hex: str) -> str:
        # No theme lookup anymore, just return the value as-is
        return name_or_hex



class Page:
    def __init__(self, bg_color=None):
        self.blocks = []
        self.bg_color = bg_color

    def add_block(self, block):
        self.blocks.append(block)

    def get_renderables(self, context=None):
        story = []
        for b in self.blocks:
            r = b.render(context)
            if isinstance(r, list):
                story.extend(r)
            else:
                story.append(r)
        return story



from abc import ABC, abstractmethod
from typing import List, Tuple, Optional, Union, Any
from reportlab.lib.colors import HexColor

from gen_pdf_abc_xyz_will_change_later.styles.context import RenderContext

class Block(ABC):
    def __init__(self, style: Optional[Any] = None):
        self.style = style

    @abstractmethod
    def render(self, context: Optional[RenderContext] = None) -> Any:
        pass

class BaseTableBlock(Block):
    def __init__(self, style: Optional[Any] = None):
        super().__init__(style=style)

    def _get_base_style(self, context: Optional['RenderContext'] = None) -> List[Tuple]:
        style_cmds = [
            ('VALIGN', (0, 0), (-1, -1), 'TOP'),
        ]
        
        # Pull visual properties from the style object
        padding = 0
        bg_color = None
        
        if self.style:
            padding = getattr(self.style, 'padding', 0) or 0
            bg_color = getattr(self.style, 'bg_color', None)
        
        style_cmds.extend([
            ('LEFTPADDING', (0, 0), (-1, -1), padding),
            ('RIGHTPADDING', (0, 0), (-1, -1), padding),
            ('TOPPADDING', (0, 0), (-1, -1), padding),
            ('BOTTOMPADDING', (0, 0), (-1, -1), padding),
        ])
        
        if bg_color:
            final_bg = bg_color
            try:
                c = HexColor(final_bg) if isinstance(final_bg, str) and final_bg.startswith('#') else final_bg
                style_cmds.append(('BACKGROUND', (0,0), (-1,-1), c))
            except Exception as e:
                print(f"Warning: Invalid bg color '{final_bg}': {e}")
            
        return style_cmds



from typing import Optional, Any, TYPE_CHECKING
if TYPE_CHECKING:
    from gen_pdf_abc_xyz_will_change_later.styles.context import RenderContext
from reportlab.platypus import Paragraph
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.styles import ParagraphStyle
from .base import Block

class TextBlock(Block):
    def __init__(self, text: str, style: Optional[Any] = None, **kwargs):
        super().__init__(style=style)
        self.text = text
        self.kwargs = kwargs
        
        self._legacy_style = None
        if not style and not kwargs:
             self._legacy_style = ParagraphStyle('Normal')
        elif isinstance(style, ParagraphStyle):
             self._legacy_style = style

    def render(self, context: Optional['RenderContext'] = None):
        if context:
            theme_style = None
            resolved_style = self.style
            
            if resolved_style and not isinstance(resolved_style, ParagraphStyle):
                theme_style = resolved_style

            rl_style_name = f'Text_{id(self)}'
            
            base_rl = ParagraphStyle('Normal')
            
            kw = dict(base_rl.__dict__) 
            kw['name'] = rl_style_name
            
            if theme_style:
                if theme_style.font_name: kw['fontName'] = theme_style.font_name
                if theme_style.font_size: kw['fontSize'] = theme_style.font_size
                if theme_style.color: kw['textColor'] = theme_style.color
                if theme_style.leading: kw['leading'] = theme_style.leading
                if theme_style.alignment is not None: kw['alignment'] = theme_style.alignment
                if theme_style.space_after: kw['spaceAfter'] = theme_style.space_after
                
            if self.kwargs:
                for k, v in self.kwargs.items():
                    if k == 'textColor' and v: kw[k] = v
                    else: kw[k] = v
            
            return Paragraph(self.text, ParagraphStyle(**kw))
            
        else:
            if self._legacy_style:
                s = self._legacy_style
            else:
                s = ParagraphStyle('Normal')
                
            if self.kwargs:
                s = ParagraphStyle(f'AdHoc_{id(self)}', parent=s, **self.kwargs)
                
            return Paragraph(self.text, s)



from .icon import IconBlock

class LucideIconBlock(IconBlock):
    BASE_URL = "https://unpkg.com/lucide-static/icons/{name}.svg"

    def __init__(self, icon_name: str, size=None, color=None, style=None):
        url = self.BASE_URL.format(name=icon_name)
        super().__init__(url, size=size, color=color, is_url=True, style=style)
        self.icon_name = icon_name



from typing import Optional
from .base import Block
from reportlab.platypus import Spacer

class IconBlock(Block):
    def __init__(self, path_or_url: str, size=None, color=None, is_url=True, style=None):
        super().__init__(style=style)
        self.size = size
        self.width = size
        self.height = size

    def render(self, context=None):
        # SVG support removed. Returning Spacer.
        w = self.width if self.width else 0
        h = self.height if self.height else 0
        return Spacer(w, h)



from reportlab.platypus import Spacer
from .base import Block

class SpacerBlock(Block):
    def __init__(self, width: float, height: float):
        self.width = width
        self.height = height

    def render(self, context=None):
        return Spacer(self.width, self.height)



from typing import List, Optional
from reportlab.platypus import Table, TableStyle
from .base import Block, BaseTableBlock
from gen_pdf_abc_xyz_will_change_later.styles.context import RenderContext

class GridBlock(BaseTableBlock):
    def __init__(self, children: List[Block], cols=2, col_widths=None, style=None):
        super().__init__(style=style)
        self.children = children
        self.cols = cols
        self.col_widths = col_widths
        
        self.width = None
        self.margin = 0
        self.padding = 0
        if self.style:
             self.width = getattr(self.style, 'width', None)
             self.margin = getattr(self.style, 'margin', 0) or 0
             self.padding = getattr(self.style, 'padding', 0) or 0

    def render(self, context: Optional['RenderContext'] = None):
        if self.width is None and self.style:
             self.width = getattr(self.style, 'width', None)
             
        grid_data = []
        current_row = []
        
        for i, child in enumerate(self.children):
            r = child.render(context)
            cell = r if isinstance(r, list) else [r]
            current_row.append(cell)
            
            if len(current_row) == self.cols:
                grid_data.append(current_row)
                current_row = []
        
        if current_row:
            while len(current_row) < self.cols:
                current_row.append('')
            grid_data.append(current_row)

        widths = self.col_widths
        if self.width and not widths:
            widths = [self.width / float(self.cols)] * self.cols
        elif widths and not isinstance(widths, list):
            widths = [widths] * self.cols
            
        t = Table(grid_data, colWidths=widths)
         
        style_cmds = self._get_base_style(context)
        t.setStyle(TableStyle(style_cmds))
        t.spaceBefore = self.margin
        t.spaceAfter = self.margin
        return t



from reportlab.lib.pagesizes import letter
from reportlab.platypus import BaseDocTemplate, PageTemplate, Frame, NextPageTemplate, PageBreak
from reportlab.lib.colors import HexColor
from reportlab.lib.units import inch
from typing import List, Optional
from .page import Page


from .styles.context import RenderContext

class Document:
    def __init__(self, pagesize=letter, pages: List[Page] = None):
        self.pages = pages or []
        self.pagesize = pagesize
        self.context = RenderContext()

    def add_page(self, page: Page):
        if not isinstance(page, Page):
            raise TypeError("Expected a Page object")
        self.pages.append(page)
    
    def append(self, pages: list):
        if not isinstance(pages, list):
             raise TypeError("Expected a list of Page objects")
        for p in pages:
            self.add_page(p)

    def _make_bg_drawer(self, color, context=None):
        def draw(canvas, doc):
            if color:
                canvas.saveState()
                final_color = color
                c = HexColor(final_color) if isinstance(final_color, str) and final_color.startswith('#') else final_color
                canvas.setFillColor(c)
                w, h = self.pagesize
                canvas.rect(0, 0, w, h, fill=True, stroke=False)
                canvas.restoreState()
        return draw

    def generate(self, filename: str):
        doc = BaseDocTemplate(filename, pagesize=self.pagesize)
        
        context = self.context
            
        margin = 72
        
        frame = Frame(
            margin, margin, 
            doc.width, doc.height, 
            id='normal'
        )
        
        templates = []
        story = []
        
        for i, page in enumerate(self.pages):
            template_id = f'PageTemplate_{i}'
            
            pt = PageTemplate(
                id=template_id, 
                frames=[frame], 
                onPage=self._make_bg_drawer(page.bg_color, context)
            )
            templates.append(pt)
            
            if i == 0:
                story.extend(page.get_renderables(context))
            else:
                story.append(NextPageTemplate(template_id))
                story.append(PageBreak())
                story.extend(page.get_renderables(context))
        
        doc.addPageTemplates(templates)
        doc.build(story)

class DocumentBuilder:
    def __init__(self, pagesize=letter):
        self.pages = []
        self.pagesize = pagesize

    def add_page(self, page: Page) -> 'DocumentBuilder':
        self.pages.append(page)
        return self

    def build(self) -> Document:
        return Document(pagesize=self.pagesize, pages=self.pages)

    def generate(self, filename: str):
        self.build().generate(filename)
